//  Copyright (c) 2007, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "shader.h"
#include "geoshader.h"

const int MAX_PATH_L = 512;

typedef struct
{
	miTag fname;
	miTag render_dir;
	int frame_padding;
	int output_format;
	int openexr_comp;
	int openexr_format;
	miBoolean filter_pass;
	miBoolean contrast_all_buffers;
//	int fb_virtual;
	miBoolean mr_color_pass;
	int mr_color_format;
	miBoolean mr_z_pass;
	int mr_z_format;
	miBoolean mr_normal_pass;
	int mr_normal_format;
	miBoolean mr_motion_pass;
	int mr_motion_format;
	miBoolean mr_label_pass;
	int mr_label_format;
	miBoolean mr_coverage_pass;
	int mr_coverage_format;
	miBoolean shading_pass;
	miBoolean color_occlusion_pass;
	miBoolean diffuse_pass;
	miBoolean diffuse_shadow_pass;
	miBoolean specular_pass;
	miBoolean specular_shadow_pass;
	miBoolean occlusion_pass;
	miBoolean indirect_illum_pass;
	miBoolean ambient_pass;
	miBoolean amb_occlusion_pass;
	miBoolean reflect_pass;
	miBoolean refract_pass;
	miBoolean incandescence_pass;
	miBoolean z_pass;
	miBoolean bent_normal_pass;
	miBoolean specialC1_pass;
	miBoolean specialC2_pass;
	miBoolean specialC3_pass;
	miBoolean specialC4_pass;
	miBoolean specialC5_pass;
	miBoolean specialC6_pass;
	miBoolean specialC7_pass;
	miBoolean specialC8_pass;
	miBoolean specialC9_pass;
	miBoolean specialC10_pass;
} p_MegaTK_pass_t;

int get_frame_number(miState* state, char *padd, int pad)
{
	int frame_n = state->camera->frame;
//	sprintf(s, "%04d", i)
//	sprintf(buffer, "%i", value);
//	std::string www;
	double int_padding = 10;
	if(frame_n != 0)
	{
		double float_padding = 1000000000;
		while((int)float_padding > frame_n)
		{
			int_padding--;
			float_padding /= 10;
		}
		if(int_padding > pad)
			pad = (int)int_padding;
	}
	else
		int_padding = 1;

	sprintf(padd, "%d", frame_n);

	char result[11];
	for(int i = 0; i < 11; i++)
		result[i] = 0;

	for(int i = (int)int_padding; i < pad; i++)
		strcat(result, "0");

	strcat(result, padd);
	strcpy(padd, result);
	return 1;
}

int get_render_file_func(miState* state, char *dir, char *name, char *prefix, int pad, char *ext)
{
	char frame_num[11];
	get_frame_number(state, frame_num, pad);
	strcat(dir, "/");				//render_dir + "/"
	strcat(dir, name);				//add name
	strcat(dir, prefix);			//add prefix
	strcat(dir, frame_num);			//add frame number
	strcat(dir, ".");				//add dot
	strcat(dir, ext);				//add format
	return 1;
}

char * get_type(int out_type, int exr_bit)
{
	char *type;
	switch(out_type)
	{
		case 0:				// 0, mean float point TIF
			type = mi_mem_strdup("rgb_fp");
			break;
		case 1:				// 1, mean 32bit IFF
			type = mi_mem_strdup("rgb_fp");
			break;
		case 2:				// 2, mean OpenEXR (mi)
			type = mi_mem_strdup("rgb_h");
			break;
		case 3:				// 3, mean OpenEXR (p_OpenEXR_saver)
			if(exr_bit == 0 && 0)	//16 bit
				type = mi_mem_strdup("rgb_h");
			else				//32 bit
				type = mi_mem_strdup("rgb_fp");
			break;
		default:				//default, float point TIF
			type = mi_mem_strdup("rgb_fp");
			break;
	}
	return type;
}

char * get_ext(int out_ext)
{
	char *ext;
	switch(out_ext)
	{
		case 0:				// 0, mean float point TIF
			ext = mi_mem_strdup("tif");
			break;
		case 1:				// 1, mean 32bit IFF
			ext = mi_mem_strdup("iff");
			break;
		case 2:				// 2, mean OpenEXR (mi)
			ext = mi_mem_strdup("exr");
			break;
		case 3:				// 3, mean OpenEXR (p_OpenEXR_saver)
			ext = mi_mem_strdup("exr");
			break;
		default:				//default, float point TIF
			ext = mi_mem_strdup("tif");
			break;
	}
	return ext;
}

/*
static PixelType  getBufferPixelType(miImg_type type, int paramType)
{
   switch (type)
   {
		case miIMG_TYPE_RGBA_16:
		case miIMG_TYPE_RGB_16 :
		case miIMG_TYPE_A_16   :
		case miIMG_TYPE_S_16   :
		case miIMG_TYPE_RGBA   :
		case miIMG_TYPE_RGB    :
		case miIMG_TYPE_A      :
		case miIMG_TYPE_S      :
		case miIMG_TYPE_VTA    :
		case miIMG_TYPE_TAG    :
		case miIMG_TYPE_BIT    :
			pixelType = Imf::UINT;
			break;
		case miIMG_TYPE_RGBA_H:
		case miIMG_TYPE_RGB_H :
			pixelType = Imf::HALF;
			break;
		case miIMG_TYPE_Z       :
		case miIMG_TYPE_A_FP    :
		case miIMG_TYPE_S_FP    :
		case miIMG_TYPE_N       :
		case miIMG_TYPE_M       :
		case miIMG_TYPE_RGBA_FP :
		case miIMG_TYPE_RGB_FP  :
		case miIMG_TYPE_COVERAGE:
		case miIMG_TYPE_RGBE    :
			pixelType = Imf::FLOAT;
			break;
   }
   return pixelType;
}
*/

miTag create_output_shaders(char *buffer_a, char *buffer_b, int operation)
{
	miTag result = 0;
	if(mi_api_function_call(mi_mem_strdup("p_math_output")))
	{
		
		if(mi_api_parameter_name(mi_mem_strdup("buffer_a")))			//buffer_a
		{
			int param_size = sizeof(buffer_a);
			mi_api_parameter_value(miTYPE_STRING, buffer_a, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: buffer_a parameter not created in function p_math_output");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("buffer_b")))			//buffer_b
		{
			int param_size = sizeof(buffer_b);
			mi_api_parameter_value(miTYPE_STRING, buffer_b, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: buffer_b parameter not created in function p_math_output");
			return result;
		}


		if(mi_api_parameter_name(mi_mem_strdup("operation")))			//operation
		{
			int *default_value = &operation;
			int param_size = sizeof(miInteger);
//			mi_api_parameter_default(miTYPE_INTEGER, default_value);
			mi_api_parameter_value(miTYPE_INTEGER, default_value, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: operation parameter not created in function p_math_output");
			return result;
		}

		miTag oldtag = 0;
		result = mi_api_function_call_end(oldtag);
		return result;
	}
	else
	{
		mi_warning("p_MegaTK_pass: function p_math_output not found");
	}	
	return result;
}

miTag declare_output_shaders(void)
{
	miTag new_func_tag;
	//declare function
	miTag check_tag = mi_api_name_lookup(mi_mem_strdup("p_math_output"));
	if(check_tag == miNULLTAG)	//fucntion not declared yet, declare it
	{
		miParameter *inputParamet;
		inputParamet = mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("buffer_a"), 0);
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("buffer_b"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("operation"), 0));

		miParameter *outParamet;
		outParamet = mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("result"), 0);
		miFunction_decl *func_dec;
		func_dec = mi_api_funcdecl_begin(outParamet, mi_mem_strdup("p_math_output"), inputParamet);
		func_dec->version = 2;
		new_func_tag = mi_api_funcdecl_end();
	}
	return new_func_tag;
}

miTag create_exr_shader(char *fileName, miState* state, p_MegaTK_pass_t *paras)
{
	miTag result = miNULLTAG;
	int	compression = *mi_eval_integer(&paras->openexr_comp) + 1;
	int all_format = *mi_eval_integer(&paras->openexr_format) + 1;

	int rgbaFormat = all_format;
	int zFormat = all_format;
	int normalFormat = all_format;
	int motionFormat = all_format;
	int labelFormat = all_format;
	int coverageFormat = all_format;

	miBoolean rgbaPass = miFALSE;
	if(*mi_eval_boolean(&paras->mr_color_pass) && *mi_eval_integer(&paras->mr_color_format) == 2)
		rgbaPass = miTRUE;

	miBoolean zPass = miFALSE;
	if(*mi_eval_boolean(&paras->mr_z_pass) && *mi_eval_integer(&paras->mr_z_format) == 4)
		zPass = miTRUE;

	miBoolean normalPass = miFALSE;
	if(*mi_eval_boolean(&paras->mr_normal_pass) && *mi_eval_integer(&paras->mr_normal_format) == 4)
		normalPass = miTRUE;

	miBoolean motionPass = miFALSE;
	if(*mi_eval_boolean(&paras->mr_motion_pass) && *mi_eval_integer(&paras->mr_motion_format) == 4)
		motionPass = miTRUE;

	miBoolean labelPass = miFALSE;
	if(*mi_eval_boolean(&paras->mr_label_pass) && *mi_eval_integer(&paras->mr_label_format) == 4)
		labelPass = miTRUE;

	miBoolean coveragePass = miFALSE;
	if(*mi_eval_boolean(&paras->mr_coverage_pass) && *mi_eval_integer(&paras->mr_coverage_format) == 4)
		coveragePass = miTRUE;

	miBoolean shading_pass =			*mi_eval_boolean(&paras->shading_pass);
	miBoolean color_occlusion_pass =	*mi_eval_boolean(&paras->color_occlusion_pass);
	miBoolean diffuse_pass =			*mi_eval_boolean(&paras->diffuse_pass);
	miBoolean diffuse_shadow_pass =		*mi_eval_boolean(&paras->diffuse_shadow_pass);
	miBoolean specular_pass =			*mi_eval_boolean(&paras->specular_pass);
	miBoolean specular_shadow_pass =	*mi_eval_boolean(&paras->specular_shadow_pass);
	miBoolean occlusion_pass =			*mi_eval_boolean(&paras->occlusion_pass);
	miBoolean indirect_illum_pass =		*mi_eval_boolean(&paras->indirect_illum_pass);
	miBoolean ambient_pass =			*mi_eval_boolean(&paras->ambient_pass);
	miBoolean amb_occlusion_pass =		*mi_eval_boolean(&paras->amb_occlusion_pass);
	miBoolean reflect_pass =			*mi_eval_boolean(&paras->reflect_pass);
	miBoolean refract_pass =			*mi_eval_boolean(&paras->refract_pass);
	miBoolean incandescence_pass =		*mi_eval_boolean(&paras->incandescence_pass);
	miBoolean z_pass =					*mi_eval_boolean(&paras->z_pass);
	miBoolean bent_normal_pass =		*mi_eval_boolean(&paras->bent_normal_pass);
	miBoolean specialC1_pass =			*mi_eval_boolean(&paras->specialC1_pass);
	miBoolean specialC2_pass =			*mi_eval_boolean(&paras->specialC2_pass);
	miBoolean specialC3_pass =			*mi_eval_boolean(&paras->specialC3_pass);
	miBoolean specialC4_pass =			*mi_eval_boolean(&paras->specialC4_pass);
	miBoolean specialC5_pass =			*mi_eval_boolean(&paras->specialC5_pass);
	miBoolean specialC6_pass =			*mi_eval_boolean(&paras->specialC6_pass);
	miBoolean specialC7_pass =			*mi_eval_boolean(&paras->specialC7_pass);
	miBoolean specialC8_pass =			*mi_eval_boolean(&paras->specialC8_pass);
	miBoolean specialC9_pass =			*mi_eval_boolean(&paras->specialC9_pass);
	miBoolean specialC10_pass =			*mi_eval_boolean(&paras->specialC10_pass);

//	miTag oldtag = miNULLTAG;
//	oldtag = mi_api_function_assign(mi_mem_strdup("p_OpenEXR_saver_from_geometry"));
//	oldtag = mi_api_decl_lookup(mi_mem_strdup("p_OpenEXR_saver_from_geometry"));
//	oldtag = mi_api_name_lookup(mi_mem_strdup("p_OpenEXR_saver_from_geometry"));
	
//	if(oldtag != 0)
//		mi_warning("OldTag is %i", int(oldtag));
//	state->camera->output

	if(mi_api_function_call(mi_mem_strdup("p_OpenEXR_saver")))
	{
		if(mi_api_parameter_name(mi_mem_strdup("fileName")))			//fileName
		{
			int param_size = sizeof(fileName);
			mi_api_parameter_value(miTYPE_STRING, fileName, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'fileName' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("compression")))			//compression
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &compression, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'compression' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("rgbaPass")))			//rgbaPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &rgbaPass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'rgbaPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("rgbaFormat")))			//rgbaFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &rgbaFormat, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'rgbaFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("zPass")))			//zPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &zPass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'zPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("zFormat")))			//zFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &zFormat, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'zFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("nPass")))			//normalPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &normalPass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'nPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("nFormat")))			//normalFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &normalFormat, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'nFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("mPass")))			//motionPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &motionPass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'mPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("mFormat")))			//motionFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &motionFormat, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'mFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("tPass")))			//labelPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &labelPass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'tPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("tFormat")))			//labelFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &labelFormat, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'tFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("cPass")))			//coveragePass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &coveragePass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'cPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("cFormat")))			//coverageFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &coverageFormat, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'cFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}


		if(*mi_eval_integer(&paras->output_format) != 3)	// output format for p_MegaTK passes is not single OpenEXR, disable it for exr saver
		{
			shading_pass =			miFALSE;
			color_occlusion_pass =	miFALSE;
			diffuse_pass =			miFALSE;
			diffuse_shadow_pass =	miFALSE;
			specular_pass =			miFALSE;
			specular_shadow_pass =	miFALSE;
			occlusion_pass =		miFALSE;
			indirect_illum_pass =	miFALSE;
			ambient_pass =			miFALSE;
			amb_occlusion_pass =	miFALSE;
			reflect_pass =			miFALSE;
			refract_pass =			miFALSE;
			incandescence_pass =	miFALSE;
			z_pass =				miFALSE;
			bent_normal_pass =		miFALSE;
			specialC1_pass =		miFALSE;
			specialC2_pass =		miFALSE;
			specialC3_pass =		miFALSE;
			specialC4_pass =		miFALSE;
			specialC5_pass =		miFALSE;
			specialC6_pass =		miFALSE;
			specialC7_pass =		miFALSE;
			specialC8_pass =		miFALSE;
			specialC9_pass =		miFALSE;
			specialC10_pass =		miFALSE;
		}

		if(mi_api_parameter_name(mi_mem_strdup("shadingPass")))				//shadingPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &shading_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'shadingPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("shadingBuffer")))			//shadingBuffer
		{
			char *name = mi_mem_strdup("tk_shading");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'shadingBuffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("shadingFormat")))			//shadingFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'shadingFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("colorPass")))				//colorPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &color_occlusion_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'colorPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("colorBuffer")))			//colorBuffer
		{
			char *name = mi_mem_strdup("tk_color");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'colorBuffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("colorFormat")))			//colorFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'colorFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("diffusePass")))				//diffusePass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &diffuse_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'diffusePass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("diffuseBuffer")))			//diffuseBuffer
		{
			char *name = mi_mem_strdup("tk_diffuse");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'diffuseBuffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("diffuseFormat")))			//diffuseFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'diffuseFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}


		if(mi_api_parameter_name(mi_mem_strdup("diffuseShadowPass")))				//diffuseShadowPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &diffuse_shadow_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'diffuseShadowPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("diffuseShadowBuffer")))			//diffuseShadowBuffer
		{
			char *name = mi_mem_strdup("tk_shadow");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'diffuseShadowBuffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("diffuseShadowFormat")))			//diffuseShadowFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'diffuseShadowFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specularPass")))				//specularPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &specular_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specularPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specularBuffer")))			//specularBuffer
		{			
			if(*mi_eval_boolean(&paras->specular_shadow_pass))
			{
				char *name = mi_mem_strdup("tk_specular");
				int param_size = sizeof(name);
				mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);	
			}
			else	//if only specular pass enabled we take shadowed specular from specularShd
			{
				char *name = mi_mem_strdup("tk_spec_shd");
				int param_size = sizeof(name);
				mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);		
			}
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specularBuffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specularFormat")))			//specularFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specularFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specularShadowPass")))				//specularShadowPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &specular_shadow_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specularShadowPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specularShadowBuffer")))			//specularShadowBuffer
		{
			char *name = mi_mem_strdup("tk_spec_shd");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specularShadowBuffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specularShadowFormat")))			//specularShadowFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specularShadowFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("diffuseOcclusionPass")))				//diffuseOcclusionPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &occlusion_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'diffuseOcclusionPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("diffuseOcclusionBuffer")))			//diffuseOcclusionBuffer
		{
			char *name = mi_mem_strdup("tk_diff_occl");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'diffuseOcclusionBuffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("diffuseOcclusionFormat")))			//diffuseOcclusionFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'diffuseOcclusionFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("indirectIllumPass")))				//indirectIllumPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &indirect_illum_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'indirectIllumPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("indirectIllumBuffer")))			//indirectIllumBuffer
		{
			char *name = mi_mem_strdup("tk_indirect");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'indirectIllumBuffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("indirectIllumFormat")))			//indirectIllumFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'indirectIllumFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}
		if(mi_api_parameter_name(mi_mem_strdup("ambientPass")))				//ambientPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &ambient_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'ambientPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("ambientBuffer")))			//ambientBuffer
		{
			char *name = mi_mem_strdup("tk_ambient");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'ambientBuffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("ambientFormat")))			//ambientFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'ambientFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("ambOcclusionPass")))				//ambOcclusionPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &amb_occlusion_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'ambOcclusionPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("ambOcclusionBuffer")))			//ambOcclusionBuffer
		{
			char *name = mi_mem_strdup("tk_amb_occlus");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'ambOcclusionBuffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("ambOcclusionFormat")))			//ambOcclusionFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'ambOcclusionFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("reflectPass")))				//reflectPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &reflect_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'reflectPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("reflectBuffer")))			//reflectBuffer
		{
			char *name = mi_mem_strdup("tk_reflect");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'reflectBuffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("reflectFormat")))			//reflectFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'reflectFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("refractPass")))				//refractPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &refract_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'refractPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("refractBuffer")))			//refractBuffer
		{
			char *name = mi_mem_strdup("tk_refract");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'refractBuffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("refractFormat")))			//refractFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'refractFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("incandescencePass")))				//incandescencePass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &incandescence_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'incandescencePass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("incandescenceBuffer")))			//incandescenceBuffer
		{
			char *name = mi_mem_strdup("tk_incand");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'incandescenceBuffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("incandescenceFormat")))			//incandescenceFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'incandescenceFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("depthPass")))				//depthPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &z_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'depthPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("depthBuffer")))			//depthBuffer
		{
			char *name = mi_mem_strdup("tk_z");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'depthBuffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("depthFormat")))			//depthFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'depthFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("bentNormalPass")))				//bentNormalPass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &bent_normal_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'bentNormalPass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("bentNormalBuffer")))			//bentNormalBuffer
		{
			char *name = mi_mem_strdup("tk_bent");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'bentNormalBuffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("bentNormalFormat")))			//bentNormalFormat
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'bentNormalFormat' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC1Pass")))				//specialC1Pass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &specialC1_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC1Pass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC1Buffer")))			//specialC1Buffer
		{
			char *name = mi_mem_strdup("tk_specialC1");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC1Buffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC1Format")))			//specialC1Format
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC1Format' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC2Pass")))				//specialC2Pass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &specialC2_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC2Pass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC2Buffer")))			//specialC2Buffer
		{
			char *name = mi_mem_strdup("tk_specialC2");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC2Buffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC2Format")))			//specialC2Format
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC2Format' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC3Pass")))				//specialC3Pass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &specialC3_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC3Pass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC3Buffer")))			//specialC3Buffer
		{
			char *name = mi_mem_strdup("tk_specialC3");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC3Buffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC3Format")))			//specialC3Format
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC3Format' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC4Pass")))				//specialC4Pass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &specialC4_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC4Pass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC4Buffer")))			//specialC4Buffer
		{
			char *name = mi_mem_strdup("tk_specialC4");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC4Buffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC4Format")))			//specialC4Format
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC4Format' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC5Pass")))				//specialC5Pass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &specialC5_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC5Pass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC5Buffer")))			//specialC5Buffer
		{
			char *name = mi_mem_strdup("tk_specialC5");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC5Buffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC5Format")))			//specialC5Format
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC5Format' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC6Pass")))				//specialC6Pass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &specialC6_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC6Pass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC6Buffer")))			//specialC6Buffer
		{
			char *name = mi_mem_strdup("tk_specialC6");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC6Buffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC6Format")))			//specialC6Format
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC6Format' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC7Pass")))				//specialC7Pass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &specialC7_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC7Pass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC7Buffer")))			//specialC7Buffer
		{
			char *name = mi_mem_strdup("tk_specialC7");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC7Buffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC7Format")))			//specialC7Format
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC7Format' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC8Pass")))				//specialC8Pass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &specialC8_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC8Pass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC8Buffer")))			//specialC8Buffer
		{
			char *name = mi_mem_strdup("tk_specialC8");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC8Buffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC8Format")))			//specialC8Format
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC8Format' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC9Pass")))				//specialC9Pass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &specialC9_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC9Pass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC9Buffer")))			//specialC9Buffer
		{
			char *name = mi_mem_strdup("tk_specialC9");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC9Buffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC9Format")))			//specialC9Format
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC9Format' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC10Pass")))				//specialC10Pass
		{
			int param_size = sizeof(miBoolean);
			mi_api_parameter_value(miTYPE_BOOLEAN, &specialC10_pass, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC10Pass' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC10Buffer")))			//specialC10Buffer
		{
			char *name = mi_mem_strdup("tk_specialC10");
			int param_size = sizeof(name);
			mi_api_parameter_value(miTYPE_STRING, name, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC10Buffer' parameter not created in function p_OpenEXR_saver");
			return result;
		}

		if(mi_api_parameter_name(mi_mem_strdup("specialC10Format")))			//specialC10Format
		{
			int param_size = sizeof(miInteger);
			mi_api_parameter_value(miTYPE_INTEGER, &all_format, NULL, &param_size);
		}
		else
		{
			mi_warning("p_MegaTK_pass: 'specialC10Format' parameter not created in function p_OpenEXR_saver");
			return result;
		}


		result = mi_api_function_call_end(miNULLTAG);
	//	if(mi_api_shader_add(mi_mem_strdup("p_OpenEXR_saver_from_geometry"), result))
	//		mi_warning("Adding name to p_OpenEXR_saver");
		return result;
	}
	else
		mi_warning("p_MegaTK_pass: function p_OpenEXR_saver was not found");

	return result;
}

miTag declare_exr_shader(void)
{
	miTag new_func_tag;
	//declare function
//	miTag check_tag = mi_api_name_lookup(mi_mem_strdup("p_OpenEXR_saver"));
	miTag check_tag = mi_api_decl_lookup(mi_mem_strdup("p_OpenEXR_saver"));
	if(check_tag == 0)	//fucntion not declared yet, declare it
	{
		miParameter *inputParamet;
		inputParamet = mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("fileName"), 0);
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("compression"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("rgbaPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("rgbaFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("zPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("zFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("nPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("nFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("mPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("mFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("tPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("tFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("cPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("cFormat"), 0));

		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("shadingPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("shadingBuffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("shadingFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("colorPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("colorBuffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("colorFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("diffusePass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("diffuseBuffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("diffuseFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("diffuseShadowPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("diffuseShadowBuffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("diffuseShadowFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("specularPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("specularBuffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("specularFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("specularShadowPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("specularShadowBuffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("specularShadowFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("diffuseOcclusionPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("diffuseOcclusionBuffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("diffuseOcclusionFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("indirectIllumPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("indirectIllumBuffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("indirectIllumFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("ambientPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("ambientBuffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("ambientFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("ambOcclusionPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("ambOcclusionBuffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("ambOcclusionFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("reflectPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("reflectBuffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("reflectFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("refractPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("refractBuffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("refractFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("incandescencePass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("incandescenceBuffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("incandescenceFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("depthPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("depthBuffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("depthFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("bentNormalPass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("bentNormalBuffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("bentNormalFormat"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("specialC1Pass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("specialC1Buffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("specialC1Format"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("specialC2Pass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("specialC2Buffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("specialC2Format"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("specialC3Pass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("specialC3Buffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("specialC3Format"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("specialC4Pass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("specialC4Buffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("specialC4Format"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("specialC5Pass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("specialC5Buffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("specialC5Format"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("specialC6Pass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("specialC6Buffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("specialC6Format"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("specialC7Pass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("specialC7Buffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("specialC7Format"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("specialC8Pass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("specialC8Buffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("specialC8Format"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("specialC9Pass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("specialC9Buffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("specialC9Format"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("specialC10Pass"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_STRING, mi_mem_strdup("specialC10Buffer"), 0));
		inputParamet = mi_api_parameter_append(inputParamet, mi_api_parameter_decl(miTYPE_INTEGER, mi_mem_strdup("specialC10Format"), 0));

		miParameter *outParamet;
		outParamet = mi_api_parameter_decl(miTYPE_BOOLEAN, mi_mem_strdup("result"), 0);
		miFunction_decl *func_dec;
		func_dec = mi_api_funcdecl_begin(outParamet, mi_mem_strdup("p_OpenEXR_saver"), inputParamet);
		func_dec->version = 3;
		new_func_tag = mi_api_funcdecl_end();
	}
	return new_func_tag;
}



extern "C"
{
DLLEXPORT int p_MegaTK_pass_version(void) {return(1);}

DLLEXPORT miBoolean p_MegaTK_pass(miTag *result, miState *state, p_MegaTK_pass_t *paras)
{
	miTag name_tag = *mi_eval_tag(&paras->fname);				//file name
	if (!name_tag)
		return miFALSE;
	char *name = (char*)mi_db_access(name_tag);
	mi_db_unpin(name_tag);

	miTag render_dir_tag = *mi_eval_tag(&paras->render_dir);	//render dir
	if (!render_dir_tag)
		return miFALSE;
	char *render_dir = (char*)mi_db_access(render_dir_tag);
	mi_db_unpin(render_dir_tag);

	int frame_padding = *mi_eval_integer(&paras->frame_padding);

	miBoolean filtering = *mi_eval_boolean(&paras->filter_pass);
	miBoolean contrast_all = *mi_eval_boolean(&paras->contrast_all_buffers);

	miCamera *camera = (miCamera *)state->camera;

	miUint typemap = miIMG_TYPE_RGB_FP;
	miUint interpmap = miIMG_TYPE_RGB_FP;

	int openexr_format = *mi_eval_integer(&paras->openexr_format);
	int out_format = *mi_eval_integer(&paras->output_format);
	int openxer_comp = *mi_eval_integer(&paras->openexr_comp);

	switch(out_format)
	{
		case 0:				// 0, mean float point TIF
			typemap = miIMG_TYPE_RGB_FP;
			interpmap = miIMG_TYPE_RGB_FP;
			break;
		case 1:				// 1, mean 32bit IFF
			typemap = miIMG_TYPE_RGB_FP;
			interpmap = miIMG_TYPE_RGB_FP;
			break;
		case 2:				// 2, mean OpenEXR (mi)
			typemap = miIMG_TYPE_RGB_H;
			interpmap = miIMG_TYPE_RGB_H;
			break;
		case 3:				// 3, mean OpenEXR (p_OpenEXR_saver)
			if(openexr_format == 0 && 0)	//16 bit (broken with gamma 3.8.1.20)
			{
				typemap = miIMG_TYPE_RGB_H;
				interpmap = miIMG_TYPE_RGB_H;
			}
			else					//32 bit
			{
				typemap = miIMG_TYPE_RGB_FP;
				interpmap = miIMG_TYPE_RGB_FP;
			}
			break;
		default:				//default, float point TIF
			typemap = miIMG_TYPE_RGB_FP;
			interpmap = miIMG_TYPE_RGB_FP;
			break;
	}

	char* exr_comp;
	switch(openxer_comp)
	{
		case 0:
			exr_comp = "none";
			break;
		case 1:
			exr_comp = "rle";
		case 2:
			exr_comp = "zip";
			break;
		case 3:
			exr_comp = "zip";
			break;
		case 4:
			exr_comp = "piz";
			break;
		case 5:
			exr_comp = "pxr24";
			break;
		default:
			exr_comp = "rle";
			break;
	}

	
	if(1)    //enable contrast all buffers
	{
		mi::shader::Interface *iface = mi_get_shader_interface();
		mi::shader::Options *options = iface->getOptions(state->options->string_options);
		iface->release();
		if(contrast_all)
			options->set("contrast all buffers", true);
		else
			options->set("contrast all buffers", false);
		options->release();
	}


	if(1)    //enable buffers in trasparency for rasterizer
	{
		mi::shader::Interface *iface = mi_get_shader_interface();
		mi::shader::Options *options = iface->getOptions(state->options->string_options);
		iface->release();
//		options->set("useopacity", true);
		options->set("rast useopacity", true);
		options->release();
	}

/////////////////////////////// Call Output Shaders
//	mi_api_incremental(1);
	mi_api_function_delete(&camera->output); //delete all output shaders

	if(*mi_eval_boolean(&paras->occlusion_pass) || *mi_eval_boolean(&paras->color_occlusion_pass))
	{
		declare_output_shaders();
		miTag out_tag = create_output_shaders(mi_mem_strdup("tk_color"), mi_mem_strdup("tk_diff_occl"), 1);
//		miTag out_tag = create_output_shaders(mi_mem_strdup("tk_diff_occl"), mi_mem_strdup("tk_color"), 1);
		miTag out_func = mi_api_output_function_def(typemap, interpmap, out_tag);
		camera->output = mi_api_function_append(camera->output, out_func);
	}
	if(*mi_eval_boolean(&paras->diffuse_pass) || *mi_eval_boolean(&paras->diffuse_shadow_pass))
	{
		declare_output_shaders();
		miTag out_tag = create_output_shaders(mi_mem_strdup("tk_shadow"), mi_mem_strdup("tk_diffuse"), 1);
		miTag out_func = mi_api_output_function_def(typemap, interpmap, out_tag);
		camera->output = mi_api_function_append(camera->output, out_func);
	}
	if(*mi_eval_boolean(&paras->specular_shadow_pass))
	{
		declare_output_shaders();
		miTag out_tag = create_output_shaders(mi_mem_strdup("tk_spec_shd"), mi_mem_strdup("tk_specular"), 1);
		miTag out_func = mi_api_output_function_def(typemap, interpmap, out_tag);
		camera->output = mi_api_function_append(camera->output, out_func);
	}

	if(*mi_eval_boolean(&paras->ambient_pass) || *mi_eval_boolean(&paras->amb_occlusion_pass))
	{
		declare_output_shaders();
		miTag out_tag = create_output_shaders(mi_mem_strdup("tk_amb_occlus"), mi_mem_strdup("tk_ambient"), 1);
		miTag out_func = mi_api_output_function_def(typemap, interpmap, out_tag);
		camera->output = mi_api_function_append(camera->output, out_func);
	}


	if(out_format == 3 ||
		(*mi_eval_boolean(&paras->mr_color_pass) && *mi_eval_integer(&paras->mr_color_format) == 2) ||
		(*mi_eval_boolean(&paras->mr_z_pass) && *mi_eval_integer(&paras->mr_z_format) == 4) ||
		(*mi_eval_boolean(&paras->mr_normal_pass) && *mi_eval_integer(&paras->mr_normal_format) == 4) ||
		(*mi_eval_boolean(&paras->mr_motion_pass) && *mi_eval_integer(&paras->mr_motion_format) == 4) ||
		(*mi_eval_boolean(&paras->mr_label_pass) && *mi_eval_integer(&paras->mr_label_format) == 4) ||
		(*mi_eval_boolean(&paras->mr_coverage_pass) && *mi_eval_integer(&paras->mr_coverage_format) == 4))		//write single OpenEXR file
	{
		char *format = get_ext(3);		//should be exr
		char *pass_prefix = mi_mem_strdup(".singleTK.");
		char temp_render_dir_cur_pass[MAX_PATH_L];
		strcpy(temp_render_dir_cur_pass, render_dir);
		get_render_file_func(state, temp_render_dir_cur_pass, name, pass_prefix, frame_padding, format);
		char *render_dir_cur_pass = mi_mem_strdup(temp_render_dir_cur_pass);

		char format_buff[40];
		format_buff[0] = 0;
		if(*mi_eval_boolean(&paras->mr_z_pass) && *mi_eval_integer(&paras->mr_z_format) == 4)
			strcat(format_buff, "+z,");
		if(*mi_eval_boolean(&paras->mr_normal_pass) && *mi_eval_integer(&paras->mr_normal_format) == 4)
			strcat(format_buff, "+n,");
		if(*mi_eval_boolean(&paras->mr_motion_pass) && *mi_eval_integer(&paras->mr_motion_format) == 4)
			strcat(format_buff, "+m,");
		if(*mi_eval_boolean(&paras->mr_label_pass) && *mi_eval_integer(&paras->mr_label_format) == 4)
			strcat(format_buff, "+tag,");
		if(*mi_eval_boolean(&paras->mr_coverage_pass) && *mi_eval_integer(&paras->mr_coverage_format) == 4)
			strcat(format_buff, "+coverage,");
		if(*mi_eval_boolean(&paras->mr_color_pass) && *mi_eval_integer(&paras->mr_color_format) == 2)
		{
			if(openexr_format == 0 && 0)	//16 bit half float (broken with gamma 3.8.1.20)
				strcat(format_buff, "+rgba_h");
			else					//32 bit float
				strcat(format_buff, "+rgba_fp");
		}
		declare_exr_shader();
		mi_api_output_type_identify(&typemap, &interpmap, mi_mem_strdup(format_buff));
		miTag out_tag = create_exr_shader(render_dir_cur_pass, state, paras);
		miTag out_func = mi_api_output_function_def(typemap, interpmap, out_tag);
		camera->output = mi_api_function_append(camera->output, out_func);
	}
//	mi_api_incremental(0);


	///////////////////////////// Define Buffers

	mi::shader::Edit_fb fbc(state->camera->buffertag);
//	fbc->reset();
	if(*mi_eval_boolean(&paras->mr_color_pass))
	{
		if(*mi_eval_integer(&paras->mr_color_format) == 2)
		{
			fbc->reset();
//			mi_warning("exr format is %i", openexr_format);
//			if (openexr_format == 1)
				fbc->set("color", "datatype", "rgba_fp");
//			else 
//				fbc->set("color", "datatype", "rgba_h");
			fbc->set("color", "user", false);
			fbc->set("color", "primary", true);
			if(filtering == miTRUE)
				fbc->set("color", "filtering", true);
			else
				fbc->set("color", "filtering", false);
		}
		else if(*mi_eval_integer(&paras->mr_color_format) == 1)
		{
			fbc->reset();
			char render_dir_cur_pass[MAX_PATH_L];
			strcpy(render_dir_cur_pass, render_dir);
			get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("color", "filename", render_dir_cur_pass);
			fbc->set("color", "filetype", "exr");
			if (openexr_format == 1)
				fbc->set("color", "datatype", "rgba_fp");
			else 
				fbc->set("color", "datatype", "rgba_h");
			fbc->set("color", "compression", exr_comp);
			fbc->set("color", "user", false);
			fbc->set("color", "primary", true);
			if(filtering == miTRUE)
				fbc->set("color", "filtering", true);
			else
				fbc->set("color", "filtering", false);
		}
	}

	if(*mi_eval_boolean(&paras->mr_z_pass))
	{
		char *z_format;
		switch(*mi_eval_integer(&paras->mr_z_format))
		{
			case 0:
				z_format = mi_mem_strdup("zt");
				break;
			case 1:
				z_format = mi_mem_strdup("map");
				break;
			case 2:
				z_format = mi_mem_strdup("tif");
				break;
			case 3:
				z_format = mi_mem_strdup("exr");
				break;
			default:
				z_format = mi_mem_strdup("zt");
				break;
		}		
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(*mi_eval_integer(&paras->mr_z_format) != 4)  //not single OpenEXR
		{
			if(*mi_eval_integer(&paras->mr_z_format) != 3)
				get_render_file_func(state, render_dir_cur_pass, name, ".depth.", frame_padding, z_format);
			else 
			{
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
				fbc->set("Z", "compression", exr_comp);
			}
			fbc->set("Z", "filename", render_dir_cur_pass);
			fbc->set("Z", "filetype", z_format);
		}
		fbc->set("Z", "datatype", "z");
		fbc->set("Z", "user", false);
		if(filtering == miTRUE)
			fbc->set("Z", "filtering", true);
		else
			fbc->set("Z", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->mr_normal_pass))
	{
		char *n_format;
		switch(*mi_eval_integer(&paras->mr_normal_format))
		{
			case 0:
				n_format = mi_mem_strdup("nt");
				break;
			case 1:
				n_format = mi_mem_strdup("map");
				break;
			case 2:
				n_format = mi_mem_strdup("tif");
				break;
			case 3:
				n_format = mi_mem_strdup("exr");
				break;
			default:
				n_format = mi_mem_strdup("nt");
				break;
		}
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(*mi_eval_integer(&paras->mr_normal_format) != 4)  //not single OpenEXR
		{
			if(*mi_eval_integer(&paras->mr_normal_format) != 3)
				get_render_file_func(state, render_dir_cur_pass, name, ".Normal.", frame_padding, n_format);
			else
			{
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
				fbc->set("Normal", "compression", exr_comp);
			}
			fbc->set("Normal", "filename", render_dir_cur_pass);
			fbc->set("Normal", "filetype", n_format);
		}
		fbc->set("Normal", "datatype", "n");
		fbc->set("Normal", "user", false);
		if(filtering == miTRUE)
			fbc->set("Normal", "filtering", true);
		else
			fbc->set("Normal", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->mr_motion_pass))
	{
		char *m_format;
		switch(*mi_eval_integer(&paras->mr_motion_format))
		{
			case 0:
				m_format = mi_mem_strdup("mt");
				break;
			case 1:
				m_format = mi_mem_strdup("map");
				break;
			case 2:
				m_format = mi_mem_strdup("tif");
				break;
			case 3:
				m_format = mi_mem_strdup("exr");
				break;
			default:
				m_format = mi_mem_strdup("mt");
				break;
		}
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(*mi_eval_integer(&paras->mr_motion_format) != 4)  //not single OpenEXR
		{
			if(*mi_eval_integer(&paras->mr_motion_format) != 3)
				get_render_file_func(state, render_dir_cur_pass, name, ".Motion.", frame_padding, m_format);
			else
			{
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
				fbc->set("Motion", "compression", exr_comp);
			}
			fbc->set("Motion", "filename", render_dir_cur_pass);
			fbc->set("Motion", "filetype", m_format);
		}
		fbc->set("Motion", "datatype", "m");
		fbc->set("Motion", "user", false);
		if(filtering == miTRUE)
			fbc->set("Motion", "filtering", true);
		else
			fbc->set("Motion", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->mr_label_pass))
	{
		char *label_format;
		switch(*mi_eval_integer(&paras->mr_label_format))
		{
			case 0:
				label_format = mi_mem_strdup("tt");
				break;
			case 1:
				label_format = mi_mem_strdup("map");
				break;
			case 2:
				label_format = mi_mem_strdup("tif");
				break;
			case 3:
				label_format = mi_mem_strdup("exr");
				break;
			default:
				label_format = mi_mem_strdup("tt");
				break;
		}
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(*mi_eval_integer(&paras->mr_label_format) != 4)  //not single OpenEXR
		{
			if(*mi_eval_integer(&paras->mr_label_format) != 3)
				get_render_file_func(state, render_dir_cur_pass, name, ".Label.", frame_padding, label_format);
			else
			{
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
				fbc->set("Label", "compression", exr_comp);
			}
			fbc->set("Label", "filename", render_dir_cur_pass);
			fbc->set("Label", "filetype", label_format);
		}
		fbc->set("Label", "datatype", "tag");
		fbc->set("Label", "user", false);
	}

	if(*mi_eval_boolean(&paras->mr_coverage_pass))
	{
		char *cov_format;
		switch(*mi_eval_integer(&paras->mr_coverage_format))
		{
			case 0:
				cov_format = mi_mem_strdup("st");
				break;
			case 1:
				cov_format = mi_mem_strdup("map");
				break;
			case 2:
				cov_format = mi_mem_strdup("tif");
				break;
			case 3:
				cov_format = mi_mem_strdup("exr");
				break;
			case 5:
				cov_format = mi_mem_strdup("null");
				break;
			default:
				cov_format = mi_mem_strdup("st");
				break;
		}
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(*mi_eval_integer(&paras->mr_coverage_format) != 4)  //not single OpenEXR
		{
			if(*mi_eval_integer(&paras->mr_coverage_format) != 3)
				get_render_file_func(state, render_dir_cur_pass, name, ".Coverage.", frame_padding, cov_format);
			else
			{
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
				fbc->set("Coverage", "compression", exr_comp);
			}
			fbc->set("Coverage", "filename", render_dir_cur_pass);
			fbc->set("Coverage", "filetype", cov_format);
		}
		fbc->set("Ccoverage", "datatype", "coverage");
		fbc->set("Coverage", "user", false);
		if(filtering == miTRUE)
			fbc->set("Ccoverage", "filtering", true);
		else
			fbc->set("Coverage", "filtering", false);
	}

	//trick for avoid using frame byffer 0, because it already used by maya glow
//	fbc->set("maya_glow_trick", "datatype", "rgb");
//	fbc->set("maya_glow_trick", "user", true);
//	fbc->set("maya_glow_trick", "filtering", true);

	if(*mi_eval_boolean(&paras->shading_pass))						//shading pass
	{
	//	fbc->remove("tk_shading");
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".shading.", frame_padding, get_ext(out_format));
			else
			{
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
				fbc->set("tk_shading", "compression", exr_comp);
			}
			fbc->set("tk_shading", "filename", render_dir_cur_pass);
			fbc->set("tk_shading", "filetype", get_ext(out_format));
		}
		fbc->set("tk_shading", "datatype", get_type(out_format, openexr_format));
//		fbc->set("tk_shading", "compression", "piz");
		fbc->set("tk_shading", "user", true);
		fbc->set("tk_shading", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_shading", "filtering", true);
		else
			fbc->set("tk_shading", "filtering", false);
//		fbc->set("tk_shading", "autofill", false);
//		fbc->set("tk_shading", "userbuffer_index", 20);
	}

	if(*mi_eval_boolean(&paras->occlusion_pass) || *mi_eval_boolean(&paras->color_occlusion_pass))				//color_occlusion pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".color.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_color", "filename", render_dir_cur_pass);
			fbc->set("tk_color", "filetype", get_ext(out_format));
		}
		fbc->set("tk_color", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_color", "user", true);
		fbc->set("tk_color", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_color", "filtering", true);
		else
			fbc->set("tk_color", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->diffuse_pass) || *mi_eval_boolean(&paras->diffuse_shadow_pass))				//diffuse pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".diffuse.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_diffuse", "filename", render_dir_cur_pass);
			fbc->set("tk_diffuse", "filetype", get_ext(out_format));
		}
		fbc->set("tk_diffuse", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_diffuse", "user", true);
		fbc->set("tk_diffuse", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_diffuse", "filtering", true);
		else
			fbc->set("tk_diffuse", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->diffuse_pass) || *mi_eval_boolean(&paras->diffuse_shadow_pass))				//shadow pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".shadow.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_shadow", "filename", render_dir_cur_pass);
			fbc->set("tk_shadow", "filetype", get_ext(out_format));
		}
		fbc->set("tk_shadow", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_shadow", "user", true);
		fbc->set("tk_shadow", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_shadow", "filtering", true);
		else
			fbc->set("tk_shadow", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->specular_shadow_pass))																	//specular_shadow pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".spec_shd.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_spec_shd", "filename", render_dir_cur_pass);
			fbc->set("tk_spec_shd", "filetype", get_ext(out_format));
		}
		fbc->set("tk_spec_shd", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_spec_shd", "user", true);
		fbc->set("tk_spec_shd", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_spec_shd", "filtering", true);
		else
			fbc->set("tk_spec_shd", "filtering", false);

		char render_dir_cur_pass1[MAX_PATH_L];
		strcpy(render_dir_cur_pass1, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass1, name, ".specular.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass1, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_specular", "filename", render_dir_cur_pass1);
			fbc->set("tk_specular", "filetype", get_ext(out_format));
		}
		fbc->set("tk_specular", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_specular", "user", true);
		fbc->set("tk_specular", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_specular", "filtering", true);
		else
			fbc->set("tk_specular", "filtering", false);

	}
	else if(*mi_eval_boolean(&paras->specular_pass))																											//specular pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".spec_shd.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_spec_shd", "filename", render_dir_cur_pass);
			fbc->set("tk_spec_shd", "filetype", get_ext(out_format));
		}
		fbc->set("tk_spec_shd", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_spec_shd", "user", true);
		fbc->set("tk_spec_shd", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_spec_shd", "filtering", true);
		else
			fbc->set("tk_spec_shd", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->occlusion_pass) || *mi_eval_boolean(&paras->color_occlusion_pass))						//occlusion pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".diff_occl.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_diff_occl", "filename", render_dir_cur_pass);
			fbc->set("tk_diff_occl", "filetype", get_ext(out_format));
		}
		fbc->set("tk_diff_occl", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_diff_occl", "user", true);
		fbc->set("tk_diff_occl", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_diff_occl", "filtering", true);
		else
			fbc->set("tk_diff_occl", "filtering", false);
	}

	
	if(*mi_eval_boolean(&paras->indirect_illum_pass))						//indirect pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".indirect.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_indirect", "filename", render_dir_cur_pass);
			fbc->set("tk_indirect", "filetype", get_ext(out_format));
		}
		fbc->set("tk_indirect", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_indirect", "user", true);
		fbc->set("tk_indirect", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_indirect", "filtering", true);
		else
			fbc->set("tk_indirect", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->ambient_pass) || *mi_eval_boolean(&paras->amb_occlusion_pass))							//ambient pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".ambient.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_ambient", "filename", render_dir_cur_pass);
			fbc->set("tk_ambient", "filetype", get_ext(out_format));
		}
		fbc->set("tk_ambient", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_ambient", "user", true);
		fbc->set("tk_ambient", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_ambient", "filtering", true);
		else
			fbc->set("tk_ambient", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->ambient_pass) || *mi_eval_boolean(&paras->amb_occlusion_pass))							//amb_occlusion pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".amb_occlus.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_amb_occlus", "filename", render_dir_cur_pass);
			fbc->set("tk_amb_occlus", "filetype", get_ext(out_format));
		}
		fbc->set("tk_amb_occlus", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_amb_occlus", "user", true);
		fbc->set("tk_amb_occlus", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_amb_occlus", "filtering", true);
		else
			fbc->set("tk_amb_occlus", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->reflect_pass))							//reflect pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".reflect.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_reflect", "filename", render_dir_cur_pass);
			fbc->set("tk_reflect", "filetype", get_ext(out_format));
		}
		fbc->set("tk_reflect", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_reflect", "user", true);
		fbc->set("tk_reflect", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_reflect", "filtering", true);
		else
			fbc->set("tk_reflect", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->refract_pass))							//refract pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".refract.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_refract", "filename", render_dir_cur_pass);
			fbc->set("tk_refract", "filetype", get_ext(out_format));
			fbc->set("tk_refract", "useopacity", true);
		}
		fbc->set("tk_refract", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_refract", "user", true);
		if(filtering == miTRUE)
			fbc->set("tk_refract", "filtering", true);
		else
			fbc->set("tk_refract", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->incandescence_pass))							//incandescence pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".incand.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_incand", "filename", render_dir_cur_pass);
			fbc->set("tk_incand", "filetype", get_ext(out_format));
		}
		fbc->set("tk_incand", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_incand", "user", true);
		fbc->set("tk_incand", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_incand", "filtering", true);
		else
			fbc->set("tk_incand", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->z_pass))									//z pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".z.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_z", "filename", render_dir_cur_pass);
			fbc->set("tk_z", "filetype", get_ext(out_format));
		}
		fbc->set("tk_z", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_z", "user", true);
		fbc->set("tk_z", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_z", "filtering", true);
		else
			fbc->set("tk_z", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->bent_normal_pass))									//bent_normal pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".bent.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_bent", "filename", render_dir_cur_pass);
			fbc->set("tk_bent", "filetype", get_ext(out_format));
		}
		fbc->set("tk_bent", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_bent", "user", true);
		fbc->set("tk_bent", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_bent", "filtering", true);
		else
			fbc->set("tk_bent", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->specialC1_pass))									//specialC1 pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".specialC1.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_specialC1", "filename", render_dir_cur_pass);
			fbc->set("tk_specialC1", "filetype", get_ext(out_format));
		}
		fbc->set("tk_specialC1", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_specialC1", "user", true);
		fbc->set("tk_specialC1", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_specialC1", "filtering", true);
		else
			fbc->set("tk_specialC1", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->specialC2_pass))									//specialC2 pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".specialC2.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_specialC2", "filename", render_dir_cur_pass);
			fbc->set("tk_specialC2", "filetype", get_ext(out_format));
		}
		fbc->set("tk_specialC2", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_specialC2", "user", true);
		fbc->set("tk_specialC2", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_specialC2", "filtering", true);
		else
			fbc->set("tk_specialC2", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->specialC3_pass))									//specialC3 pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".specialC3.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_specialC3", "filename", render_dir_cur_pass);
			fbc->set("tk_specialC3", "filetype", get_ext(out_format));
		}
		fbc->set("tk_specialC3", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_specialC3", "user", true);
		fbc->set("tk_specialC3", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_specialC3", "filtering", true);
		else
			fbc->set("tk_specialC3", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->specialC4_pass))									//specialC4 pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".specialC4.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_specialC4", "filename", render_dir_cur_pass);
			fbc->set("tk_specialC4", "filetype", get_ext(out_format));
		}
		fbc->set("tk_specialC4", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_specialC4", "user", true);
		fbc->set("tk_specialC4", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_specialC4", "filtering", true);
		else
			fbc->set("tk_specialC4", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->specialC5_pass))									//specialC5 pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".specialC5.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_specialC5", "filename", render_dir_cur_pass);
			fbc->set("tk_specialC5", "filetype", get_ext(out_format));
		}
		fbc->set("tk_specialC5", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_specialC5", "user", true);
		fbc->set("tk_specialC5", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_specialC5", "filtering", true);
		else
			fbc->set("tk_specialC5", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->specialC6_pass))									//specialC6 pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".specialC6.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_specialC6", "filename", render_dir_cur_pass);
			fbc->set("tk_specialC6", "filetype", get_ext(out_format));
		}
		fbc->set("tk_specialC6", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_specialC6", "user", true);
		fbc->set("tk_specialC6", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_specialC6", "filtering", true);
		else
			fbc->set("tk_specialC6", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->specialC7_pass))									//specialC7 pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".specialC7.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_specialC7", "filename", render_dir_cur_pass);
			fbc->set("tk_specialC7", "filetype", get_ext(out_format));
		}
		fbc->set("tk_specialC7", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_specialC7", "user", true);
		fbc->set("tk_specialC7", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_specialC7", "filtering", true);
		else
			fbc->set("tk_specialC7", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->specialC8_pass))									//specialC8 pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".specialC8.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_specialC8", "filename", render_dir_cur_pass);
			fbc->set("tk_specialC8", "filetype", get_ext(out_format));
		}
		fbc->set("tk_specialC8", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_specialC8", "user", true);
		fbc->set("tk_specialC8", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_specialC8", "filtering", true);
		else
			fbc->set("tk_specialC8", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->specialC9_pass))									//specialC9 pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".specialC9.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_specialC9", "filename", render_dir_cur_pass);
			fbc->set("tk_specialC9", "filetype", get_ext(out_format));
		}
		fbc->set("tk_specialC9", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_specialC9", "user", true);
		fbc->set("tk_specialC9", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_specialC9", "filtering", true);
		else
			fbc->set("tk_specialC9", "filtering", false);
	}

	if(*mi_eval_boolean(&paras->specialC10_pass))									//specialC10 pass
	{
		char render_dir_cur_pass[MAX_PATH_L];
		strcpy(render_dir_cur_pass, render_dir);
		if(out_format != 3)  //not single OpenEXR
		{
			if(out_format != 2)
				get_render_file_func(state, render_dir_cur_pass, name, ".specialC10.", frame_padding, get_ext(out_format));
			else 
				get_render_file_func(state, render_dir_cur_pass, name, ".single.", frame_padding, get_ext(out_format));
			fbc->set("tk_specialC10", "filename", render_dir_cur_pass);
			fbc->set("tk_specialC10", "filetype", get_ext(out_format));
		}
		fbc->set("tk_specialC10", "datatype", get_type(out_format, openexr_format));
		fbc->set("tk_specialC10", "user", true);
		fbc->set("tk_specialC10", "useopacity", true);
		if(filtering == miTRUE)
			fbc->set("tk_specialC10", "filtering", true);
		else
			fbc->set("tk_specialC10", "filtering", false);
	}



	return(miTRUE);
}
}
