//  Copyright (c) 2005, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"

typedef struct
{
	miColor		color;
	miScalar	alpha;
	miScalar	ior;
}p_color_plus_alpha_t;

#define miEPS 0.0001

extern "C"
{
DLLEXPORT int p_color_plus_alpha_version(void) {return(2);}

DLLEXPORT miBoolean p_color_plus_alpha(miColor *result, miState *state,	p_color_plus_alpha_t *paras)
{
	if(state->type == miRAY_SHADOW || state->type == miRAY_DISPLACE)
		return miFALSE;

	miScalar ior;
	miScalar alpha;

	alpha = *mi_eval_scalar(&paras->alpha);
	ior = *mi_eval_scalar(&paras->ior);
	if(alpha < miEPS)
	{
		if(ior == 1)
			mi_trace_transparent(result, state);
//			mi_trace_continue(result, state);
		else
		{
			miVector direction;
			mi_transmission_dir_specular(&direction, state, ior, 1);
			mi_trace_refraction(result, state, &direction);
		}
	}
	else
	{
		if(alpha == 1)
		{
			*result = *mi_eval_color(&paras->color);
			result->a = 1;
		}
		else
		{
			miColor *res_color;
			res_color = mi_eval_color(&paras->color);
			if(ior == 1)
				mi_trace_transparent(result, state);
			else
			{
				miVector direction;
				mi_transmission_dir_specular(&direction, state, ior, 1);
				mi_trace_refraction(result, state, &direction);
			}
			result->r = (1 - alpha) * result->r + res_color->r * alpha;
			result->g = (1 - alpha) * result->g + res_color->g * alpha;
			result->b = (1 - alpha) * result->b + res_color->b * alpha;
//			result->a = (1 - alpha) * result->a + alpha;
			result->a = result->a + alpha;
			if(result->a > 1)
				result->a = 1;
			else
			{
				if(result->a < 0)
					result->a = 0;
			}
		}
	}
	return(miTRUE);
}
}

