//  Copyright (c) 2005, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"

typedef struct
{
	miColor color;
	int	conversion;
	miScalar clamp_min;
	miScalar clamp_max;
	miBoolean invert;
}p_color_to_float_t;

extern "C"
{
DLLEXPORT int p_color_to_float_version(void) {return(2);}

DLLEXPORT miBoolean p_color_to_float(miScalar *result, miState *state, p_color_to_float_t *paras)
{
	miBoolean invert_float = *mi_eval_boolean(&paras->invert);
	miColor color = *mi_eval_color(&paras->color);
	int convers = *mi_eval_integer(&paras->conversion);

	switch(convers)
	{
		case 0:		//r
			*result = color.r;
			break;
		case 1:		//g
			*result = color.g;
			break;
		case 2:		//b
			*result = color.b;
			break;
		case 3:		//a
			*result = color.a;
			break;
		case 4:		//average
			*result = (color.r + color.g + color.b)/3;
			break;
		case 5:		//luminosity
			*result = color.r*0.299f + color.g*0.587f + color.b*0.114f;
			break;
		default:	//default average
			*result = (color.r + color.g + color.b)/3;
			break;
	}

	miScalar max = *mi_eval_scalar(&paras->clamp_max);
	miScalar min = *mi_eval_scalar(&paras->clamp_min);

	if(*result > max)
		*result = max;
	else if(*result < min)
			*result = min;

	if(invert_float == miTRUE)
		*result = 1 - *result;

	return(miTRUE);
}
}
