//  Copyright (c) 2005, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"

typedef struct
{
	miColor		color1;
	miColor		color2;
	miScalar	interpolation;
}p_facing_ratio_t;

extern "C"
{
DLLEXPORT int p_facing_ratio_version(void)
{
	return(2);
}

DLLEXPORT miBoolean p_facing_ratio(miColor *result, miState *state, p_facing_ratio_t *paras)
{
	if(state->type == miRAY_SHADOW || state->type == miRAY_DISPLACE)
		return miFALSE;

	miVector dir;
	miVector nrm;
	miColor *col1;
	miColor *col2;
	miScalar inter;

	mi_point_to_camera(state, &dir, &state->point);
	mi_vector_neg(&dir);
	mi_vector_normalize(&dir);
	mi_vector_to_camera(state, &nrm, &state->normal);
	mi_vector_normalize(&nrm);
	float r = mi_vector_dot(&dir, &nrm);

	inter = *mi_eval_scalar(&paras->interpolation);

	if(r < 0)
		r = 0;
	else if(r > 1)
		r = 1;

	if(r != 1)
	{
		r = pow(r, inter);
		if(r < 0)
			r = 0;
		else if(r > 1)
			r = 1;
	}

	col1 = mi_eval_color(&paras->color1);
	col2 = mi_eval_color(&paras->color2);

	result->r = (col1->r * r) + (col2->r * (1 - r));
	result->g = (col1->g * r) + (col2->g * (1 - r));
	result->b = (col1->b * r) + (col2->b * (1 - r));
	result->a = 1;

	return(miTRUE);
}
}
