//  Copyright (c) 2005, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"

typedef struct
{
	miScalar power;
}p_fisheye_t;

extern "C"
{
DLLEXPORT int p_fisheye_version(void) {return(2);}

DLLEXPORT miBoolean p_fisheye(miColor *result, miState *state, p_fisheye_t  *paras)
{
	if(state->type == miRAY_SHADOW || state->type == miRAY_DISPLACE)
		return miFALSE;

	miVector        camdir, dir;
	miScalar        x, y, r, t;
	miScalar		pow;
	pow = *mi_eval_scalar(&paras->power);

	if(pow == 0)
		return(mi_trace_eye(result, state, &state->org, &state->dir));

	else
	{
		mi_vector_to_camera(state, &camdir, &state->dir);
		t = state->camera->focal / -camdir.z /
                                (state->camera->aperture/2)/2;
		x = t * camdir.x;
		y = t * camdir.y * state->camera->aspect;
		r = x * x + y * y;

		dir.x = camdir.x * r;
		dir.y = camdir.y * r;
		dir.z = -sqrt(1 - dir.x*dir.x - dir.y*dir.y);

		dir.x = camdir.x + (dir.x - camdir.x)*pow;
		dir.y = camdir.y + (dir.y - camdir.y)*pow;
		dir.z = camdir.z + (dir.z - camdir.z)*pow;

		mi_vector_from_camera(state, &dir, &dir);
		return(mi_trace_eye(result, state, &state->org, &dir));
	}
}
}
