//  Copyright (c) 2007, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"

typedef struct
{
	int			mode;
	miScalar	repeat;
}p_hair_mapping_t;

extern "C"
{
DLLEXPORT int p_hair_mapping_version(void) {return(1);}

DLLEXPORT miBoolean p_hair_mapping(miScalar *result, miState *state, p_hair_mapping_t *paras)
{
	int		mode = *mi_eval_integer(&paras->mode);
	float	repeat = *mi_eval_scalar(&paras->repeat);
	float	coord = 0.f;

	switch(mode)
	{
		case 0:		//width
			if(state->bary[0] == state->bary[0])
				coord = (miScalar)state->bary[0];
			else
				mi_warning("p_hair_mapping: NaN or infinite value in state->bary[0]. Replaced with zero");
			break;
		case 1:		//length
			if(state->bary[1] == state->bary[1])
				coord = (miScalar)state->bary[1];
			else
				mi_warning("p_hair_mapping: NaN or infinite value in state->bary[1]. Replaced with zero");
			break;
		default:
			coord = 0.f;
			break;
	}

	if(repeat != 1.f && repeat)
	{
		coord *= repeat;
		while(coord > 1)
			coord--;
	}
	*result = coord;
	return(miTRUE);
}
}
