//  Copyright (c) 2007, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


// Modification of original physical_lens_dof shader from mental images
// Added samples

#include <stdio.h>
#include <math.h>
#include <shader.h>
#include "p_static_param.h"

#ifdef DEBUG
#define miASSERT assert
#else
#define miASSERT(x)
#endif

static miColor black = {0.0, 0.0, 0.0, 0.0};


struct dof {
	miScalar	distance;		/* distance to focused plane */
	miScalar	radius;		/* size of the lens */
	int			samples;	/* dof samples */
};

extern "C" DLLEXPORT miBoolean p_physical_lens_dof_init(miState *state, dof	*paras, miBoolean *inst_req)
{
	struct static_param_dof *sp, **spp;

	if (!paras)
	{
		*inst_req = miTRUE;
		return(miTRUE);
	}

	sp = (struct static_param_dof *)
	mi_mem_allocate(sizeof(struct static_param_dof));

	mi_query(miQ_FUNC_USERPTR, state, 0, &spp);
	*spp = sp;

	mi::shader::Access_fb fbc(state->camera->buffertag);
	size_t index;

	index = -1;
	if(fbc->get_index("tk_shading", index))
		sp->shading_pass = miTRUE;
	else 
		sp->shading_pass = miFALSE;
	sp->shading_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_color", index))
		sp->color_pass = miTRUE;
	else 
		sp->color_pass = miFALSE; 
	sp->color_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_diffuse", index))
		sp->diffuse_pass = miTRUE;
	else 
		sp->diffuse_pass = miFALSE;
	sp->diffuse_buf = (int)index;
	
	index = -1;
	if(fbc->get_index("tk_shadow", index))
		sp->diffuse_shadow_pass = miTRUE;
	else 
		sp->diffuse_shadow_pass = miFALSE; 
	sp->diffuse_shadow_buf  = (int)index;
	
	index = -1;
	if(fbc->get_index("tk_specular", index))
		sp->specular_pass = miTRUE;
	else 
		sp->specular_pass = miFALSE; 
	sp->specular_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_spec_shd", index))
		sp->specular_shadow_pass = miTRUE;
	else 
		sp->specular_shadow_pass = miFALSE; 
	sp->specular_shadow_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_diff_occl", index))
		sp->occlusion_pass = miTRUE;
	else 
		sp->occlusion_pass = miFALSE; 
	sp->occlusion_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_indirect", index))
		sp->indirect_illum_pass = miTRUE;
	else 
		sp->indirect_illum_pass = miFALSE; 
	sp->indirect_illum_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_ambient", index))
		sp->ambient_pass = miTRUE;
	else 
		sp->ambient_pass = miFALSE; 
	sp->ambient_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_amb_occlus", index))
		sp->amb_occlusion_pass = miTRUE;
	else 
		sp->amb_occlusion_pass = miFALSE;
	sp->amb_occlusion_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_reflect", index))
		sp->reflect_pass = miTRUE;
	else 
		sp->reflect_pass = miFALSE;
	sp->reflect_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_refract", index))
		sp->refract_pass = miTRUE;
	else 
		sp->refract_pass = miFALSE;
	sp->refract_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_incand", index))
		sp->incandescence_pass = miTRUE;
	else 
		sp->incandescence_pass = miFALSE; 
	sp->incandescence_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_z", index))
		sp->z_pass = miTRUE;
	else 
		sp->z_pass = miFALSE;
	sp->z_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_bent", index))
		sp->bent_normal_pass = miTRUE;
	else 
		sp->bent_normal_pass = miFALSE;
	sp->bent_normal_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_specialC1", index))
		sp->specialC1_pass = miTRUE;
	else 
		sp->specialC1_pass = miFALSE;
	sp->specialC1_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_specialC2", index))
		sp->specialC2_pass = miTRUE;
	else 
		sp->specialC2_pass = miFALSE; 
	sp->specialC2_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_specialC3", index))
		sp->specialC3_pass = miTRUE;
	else 
		sp->specialC3_pass = miFALSE;
	sp->specialC3_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_specialC4", index))
		sp->specialC4_pass = miTRUE;
	else 
		sp->specialC4_pass = miFALSE;
	sp->specialC4_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_specialC5", index))
		sp->specialC5_pass = miTRUE;
	else 
		sp->specialC5_pass = miFALSE;
	sp->specialC5_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_specialC6", index))
		sp->specialC6_pass = miTRUE;
	else 
		sp->specialC6_pass = miFALSE;
	sp->specialC6_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_specialC7", index))
		sp->specialC7_pass = miTRUE;
	else 
		sp->specialC7_pass = miFALSE;
	sp->specialC7_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_specialC8", index))
		sp->specialC8_pass = miTRUE;
	else 
		sp->specialC8_pass = miFALSE; 
	sp->specialC8_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_specialC9", index))
		sp->specialC9_pass = miTRUE;
	else 
		sp->specialC9_pass = miFALSE;
	sp->specialC9_buf = (int)index;

	index = -1;
	if(fbc->get_index("tk_specialC10", index))
		sp->specialC10_pass = miTRUE;
	else 
		sp->specialC10_pass = miFALSE;
	sp->specialC10_buf = (int)index;

	return(miTRUE);
}

extern "C" DLLEXPORT miBoolean p_physical_lens_dof_exit(miState *state, dof *paras)
{
	struct static_param_dof **spp;

	if (paras)
	{
		mi_query(miQ_FUNC_USERPTR, state, 0, &spp);
		mi_mem_release(*spp);
		*spp = NULL;
	}
	return(miTRUE);
}


extern "C" DLLEXPORT int p_physical_lens_dof_version(void) {return(1);}

extern "C" DLLEXPORT miBoolean p_physical_lens_dof(
	register miColor	*result,
	register miState	*state,
	register struct dof	*paras)
{
	miScalar	distance = *mi_eval_scalar(&paras->distance) * -1;
	miScalar	rads  = *mi_eval_scalar(&paras->radius);
	miVector	org, dir;	/* original state->org, state->dir */
	miVector	rayorg, raydir;	/* ray origin and direction */
	miVector	dest;
	double		t, radius, angle;
	miColor		average;
	int		i;
	double		pad[2];
	miUint		n;		/* number of eye rays to shoot */
	int		dummy[4];	/* for tile size query */
	float	inv_sam = 0;

	struct static_param_dof *sp, **spp;
	miASSERT(state->shader);
	mi_query(miQ_FUNC_USERPTR, state, 0, &spp);
	sp = *spp;

	average = black;
	mi_point_to_camera (state, &org, &state->org);
	mi_vector_to_camera(state, &dir, &state->dir);
	t = (distance - org.z) / dir.z;
	dest.x = (miScalar)(org.x + t * dir.x);
	dest.y = (miScalar)(org.y + t * dir.y);
	dest.z = distance;

	i = 0;

	/*
	 * In the finalgather precomputing stage multiple eye rays are
	 * useless. To detect that shader is called im the finalgather
	 * precomputing mode, the fact that there are no "tiles" and thus
	 * mi_query(miQ_TILE_PIXELS) returns miFALSE is used.
	 */
	if (state->options->finalgather &&
	    !mi_query(miQ_TILE_PIXELS, state, 0, dummy))
		n = 1;
	else
		n = *mi_eval_integer(&paras->samples);

	if(n > 1)
		inv_sam = 1.f/n;
	else
	{
		inv_sam = 1.f;
		n = 1;
	}

	miColor sum_shading =			black;
	miColor sum_color =				black;
	miColor sum_diffuse =			black;
	miColor sum_diffuse_shadow =	black;
	miColor sum_specular =			black;
	miColor sum_specular_shadow =	black;
	miColor sum_occlusion =			black;
	miColor sum_ambient =			black;
	miColor sum_amb_occlusion =		black;
	miColor sum_indirect_illum =	black;
	miColor sum_reflect =			black;
	miColor sum_refract =			black;
	miColor sum_incandescence =		black;
	miColor sum_z =					black;
	miColor sum_bent_normal =		black;
	miColor sum_specialC1 =			black;
	miColor sum_specialC2 =			black;
	miColor sum_specialC3 =			black;
	miColor sum_specialC4 =			black;
	miColor sum_specialC5 =			black;
	miColor sum_specialC6 =			black;
	miColor sum_specialC7 =			black;
	miColor sum_specialC8 =			black;
	miColor sum_specialC9 =			black;
	miColor sum_specialC10 =		black;

	miColor current_shading =			black;
	miColor current_color =				black;
	miColor current_diffuse =			black;
	miColor current_diffuse_shadow =	black;
	miColor current_specular =			black;
	miColor current_specular_shadow =	black;
	miColor current_occlusion =			black;
	miColor current_ambient =			black;
	miColor current_amb_occlusion =		black;
	miColor current_indirect_illum =	black;
	miColor current_reflect =			black;
	miColor current_refract =			black;
	miColor current_incandescence =		black;
	miColor current_z =					black;
	miColor current_bent_normal =		black;
	miColor current_specialC1 =			black;
	miColor current_specialC2 =			black;
	miColor current_specialC3 =			black;
	miColor current_specialC4 =			black;
	miColor current_specialC5 =			black;
	miColor current_specialC6 =			black;
	miColor current_specialC7 =			black;
	miColor current_specialC8 =			black;
	miColor current_specialC9 =			black;
	miColor current_specialC10 =		black;


	while(mi_sample(pad, &i, state, 2, &n))
	{
		radius = rads * sqrt(pad[0]);
		angle = 2.0 * M_PI * pad[1];

		rayorg.x = (miScalar)(org.x + radius * cos(angle));
		rayorg.y = (miScalar)(org.y + radius * sin(angle));
		rayorg.z = org.z;
		raydir.x = dest.x - rayorg.x;
		raydir.y = dest.y - rayorg.y;
		raydir.z = dest.z - rayorg.z;
		mi_point_from_camera (state, &rayorg, &rayorg);
		mi_vector_from_camera(state, &raydir, &raydir);

		state->pri = NULL;
		(void)mi_trace_eye(result, state, &rayorg, &raydir);

		average.r += result->r;
		average.g += result->g;
		average.b += result->b;
		average.a += result->a;


		if(1)	//Get TK passes
		{
			if(sp->shading_pass)
			{
				current_shading = black;
				if(mi_fb_get(state, sp->shading_buf, &current_shading))
				{
					sum_shading.r += current_shading.r;
					sum_shading.g += current_shading.g;
					sum_shading.b += current_shading.b;
				}
			}
			if(sp->color_pass)
			{
				current_color = black;
				if(mi_fb_get(state, sp->color_buf, &current_color))
				{
					sum_color.r += current_color.r;
					sum_color.g += current_color.g;
					sum_color.b += current_color.b;
				}
			}
			if(sp->diffuse_pass)
			{
				current_diffuse = black;
				if(mi_fb_get(state, sp->diffuse_buf, &current_diffuse))
				{
					sum_diffuse.r += current_diffuse.r;
					sum_diffuse.g += current_diffuse.g;
					sum_diffuse.b += current_diffuse.b;
				}
			}
			if(sp->diffuse_shadow_pass)
			{
				current_diffuse_shadow = black;
				if(mi_fb_get(state, sp->diffuse_shadow_buf, &current_diffuse_shadow))
				{
					sum_diffuse_shadow.r += current_diffuse_shadow.r;
					sum_diffuse_shadow.g += current_diffuse_shadow.g;
					sum_diffuse_shadow.b += current_diffuse_shadow.b;
				}
			}
			if(sp->specular_pass)
			{
				current_specular = black;
				if(mi_fb_get(state, sp->specular_buf, &current_specular))
				{
					sum_specular.r += current_specular.r;
					sum_specular.g += current_specular.g;
					sum_specular.b += current_specular.b;
				}
			}
			if(sp->specular_shadow_pass)
			{
				current_specular_shadow = black;
				if(mi_fb_get(state, sp->specular_shadow_buf, &current_specular_shadow))
				{
					sum_specular_shadow.r += current_specular_shadow.r;
					sum_specular_shadow.g += current_specular_shadow.g;
					sum_specular_shadow.b += current_specular_shadow.b;
				}
			}
			if(sp->occlusion_pass)
			{
				current_occlusion = black;
				if(mi_fb_get(state, sp->occlusion_buf, &current_occlusion))
				{
					sum_occlusion.r += current_occlusion.r;
					sum_occlusion.g += current_occlusion.g;
					sum_occlusion.b += current_occlusion.b;
				}
			}
			if(sp->ambient_pass)
			{
				current_ambient = black;
				if(mi_fb_get(state, sp->ambient_buf, &current_ambient))
				{
					sum_ambient.r += current_ambient.r;
					sum_ambient.g += current_ambient.g;
					sum_ambient.b += current_ambient.b;
				}
			}
			if(sp->amb_occlusion_pass)
			{
				current_amb_occlusion = black;
				if(mi_fb_get(state, sp->amb_occlusion_buf, &current_amb_occlusion))
				{
					sum_amb_occlusion.r += current_amb_occlusion.r;
					sum_amb_occlusion.g += current_amb_occlusion.g;
					sum_amb_occlusion.b += current_amb_occlusion.b;
				}
			}
			if(sp->indirect_illum_pass)
			{
				current_indirect_illum = black;
				if(mi_fb_get(state, sp->indirect_illum_buf, &current_indirect_illum))
				{
					sum_indirect_illum.r += current_indirect_illum.r;
					sum_indirect_illum.g += current_indirect_illum.g;
					sum_indirect_illum.b += current_indirect_illum.b;
				}
			}
			if(sp->reflect_pass)
			{
				current_reflect = black;
				if(mi_fb_get(state, sp->reflect_buf, &current_reflect))
				{
					sum_reflect.r += current_reflect.r;
					sum_reflect.g += current_reflect.g;
					sum_reflect.b += current_reflect.b;
				}
			}
			if(sp->refract_pass)
			{
				current_refract = black;
				if(mi_fb_get(state, sp->refract_buf, &current_refract))
				{
					sum_refract.r += current_refract.r;
					sum_refract.g += current_refract.g;
					sum_refract.b += current_refract.b;
				}
			}
			if(sp->incandescence_pass)
			{
				current_incandescence = black;
				if(mi_fb_get(state, sp->incandescence_buf, &current_incandescence))
				{
					sum_incandescence.r += current_incandescence.r;
					sum_incandescence.g += current_incandescence.g;
					sum_incandescence.b += current_incandescence.b;
				}
			}
			if(sp->z_pass)
			{
				current_z = black;
				if(mi_fb_get(state, sp->z_buf, &current_z))
				{
					sum_z.r += current_z.r;
					sum_z.g += current_z.g;
					sum_z.b += current_z.b;
				}
			}
			if(sp->bent_normal_pass)
			{
				current_bent_normal = black;
				if(mi_fb_get(state, sp->bent_normal_buf, &current_bent_normal))
				{
					sum_bent_normal.r += current_bent_normal.r;
					sum_bent_normal.g += current_bent_normal.g;
					sum_bent_normal.b += current_bent_normal.b;
				}
			}
			if(sp->specialC1_pass)
			{
				current_specialC1 = black;
				if(mi_fb_get(state, sp->specialC1_buf, &current_specialC1))
				{
					sum_specialC1.r += current_specialC1.r;
					sum_specialC1.g += current_specialC1.g;
					sum_specialC1.b += current_specialC1.b;
				}
			}
			if(sp->specialC2_pass)
			{
				current_specialC2 = black;
				if(mi_fb_get(state, sp->specialC2_buf, &current_specialC2))
				{
					sum_specialC2.r += current_specialC2.r;
					sum_specialC2.g += current_specialC2.g;
					sum_specialC2.b += current_specialC2.b;
				}
			}
			if(sp->specialC3_pass)
			{
				current_specialC3 = black;
				if(mi_fb_get(state, sp->specialC3_buf, &current_specialC3))
				{
					sum_specialC3.r += current_specialC3.r;
					sum_specialC3.g += current_specialC3.g;
					sum_specialC3.b += current_specialC3.b;
				}
			}
			if(sp->specialC4_pass)
			{
				current_specialC4 = black;
				if(mi_fb_get(state, sp->specialC4_buf, &current_specialC4))
				{
					sum_specialC4.r += current_specialC4.r;
					sum_specialC4.g += current_specialC4.g;
					sum_specialC4.b += current_specialC4.b;
				}
			}
			if(sp->specialC5_pass)
			{
				current_specialC5 = black;
				if(mi_fb_get(state, sp->specialC5_buf, &current_specialC5))
				{
					sum_specialC5.r += current_specialC5.r;
					sum_specialC5.g += current_specialC5.g;
					sum_specialC5.b += current_specialC5.b;
				}
			}
			if(sp->specialC6_pass)
			{
				current_specialC6 = black;
				if(mi_fb_get(state, sp->specialC6_buf, &current_specialC6))
				{
					sum_specialC6.r += current_specialC6.r;
					sum_specialC6.g += current_specialC6.g;
					sum_specialC6.b += current_specialC6.b;
				}
			}
			if(sp->specialC7_pass)
			{
				current_specialC7 = black;
				if(mi_fb_get(state, sp->specialC7_buf, &current_specialC7))
				{
					sum_specialC7.r += current_specialC7.r;
					sum_specialC7.g += current_specialC7.g;
					sum_specialC7.b += current_specialC7.b;
				}
			}
			if(sp->specialC8_pass)
			{
				current_specialC8 = black;
				if(mi_fb_get(state, sp->specialC8_buf, &current_specialC8))
				{
					sum_specialC8.r += current_specialC8.r;
					sum_specialC8.g += current_specialC8.g;
					sum_specialC8.b += current_specialC8.b;
				}
			}
			if(sp->specialC9_pass)
			{
				current_specialC9 = black;
				if(mi_fb_get(state, sp->specialC9_buf, &current_specialC9))
				{
					sum_specialC9.r += current_specialC9.r;
					sum_specialC9.g += current_specialC9.g;
					sum_specialC9.b += current_specialC9.b;
				}
			}
			if(sp->specialC10_pass)
			{
				current_specialC10 = black;
				if(mi_fb_get(state, sp->specialC10_buf, &current_specialC10))
				{
					sum_specialC10.r += current_specialC10.r;
					sum_specialC10.g += current_specialC10.g;
					sum_specialC10.b += current_specialC10.b;
				}
			}
		}
    }
	result->r = average.r * inv_sam;
	result->g = average.g * inv_sam;
	result->b = average.b * inv_sam;
	result->a = average.a * inv_sam;

	if(1)	//Put TK passes
	{
		if(sp->shading_pass)
		{
			sum_shading.r *= inv_sam;
			sum_shading.g *= inv_sam;
			sum_shading.b *= inv_sam;
			mi_fb_put(state, sp->shading_buf, &sum_shading);
		}
		if(sp->color_pass)
		{
			sum_color.r *= inv_sam;
			sum_color.g *= inv_sam;
			sum_color.b *= inv_sam;
			mi_fb_put(state, sp->color_buf, &sum_color);
		}
		if(sp->diffuse_pass)
		{
			sum_diffuse.r *= inv_sam;
			sum_diffuse.g *= inv_sam;
			sum_diffuse.b *= inv_sam;
			mi_fb_put(state, sp->diffuse_buf, &sum_diffuse);
		}
		if(sp->diffuse_shadow_pass)
		{
			sum_diffuse_shadow.r *= inv_sam;
			sum_diffuse_shadow.g *= inv_sam;
			sum_diffuse_shadow.b *= inv_sam;
			mi_fb_put(state, sp->diffuse_shadow_buf, &sum_diffuse_shadow);
		}
		if(sp->specular_pass)
		{
			sum_specular.r *= inv_sam;
			sum_specular.g *= inv_sam;
			sum_specular.b *= inv_sam;
			mi_fb_put(state, sp->specular_buf, &sum_specular);
		}
		if(sp->specular_shadow_pass)
		{
			sum_specular_shadow.r *= inv_sam;
			sum_specular_shadow.g *= inv_sam;
			sum_specular_shadow.b *= inv_sam;
			mi_fb_put(state, sp->specular_shadow_buf, &sum_specular_shadow);
		}
		if(sp->occlusion_pass)
		{
			sum_occlusion.r *= inv_sam;
			sum_occlusion.g *= inv_sam;
			sum_occlusion.b *= inv_sam;
			mi_fb_put(state, sp->occlusion_buf, &sum_occlusion);
		}
		if(sp->ambient_pass)
		{
			sum_ambient.r *= inv_sam;
			sum_ambient.g *= inv_sam;
			sum_ambient.b *= inv_sam;
			mi_fb_put(state, sp->ambient_buf, &sum_ambient);
		}
		if(sp->amb_occlusion_pass)
		{
			sum_amb_occlusion.r *= inv_sam;
			sum_amb_occlusion.g *= inv_sam;
			sum_amb_occlusion.b *= inv_sam;
			mi_fb_put(state, sp->amb_occlusion_buf, &sum_amb_occlusion);
		}
		if(sp->indirect_illum_pass)
		{
			sum_indirect_illum.r *= inv_sam;
			sum_indirect_illum.g *= inv_sam;
			sum_indirect_illum.b *= inv_sam;
			mi_fb_put(state, sp->indirect_illum_buf, &sum_indirect_illum);
		}
		if(sp->reflect_pass)
		{
			sum_reflect.r *= inv_sam;
			sum_reflect.g *= inv_sam;
			sum_reflect.b *= inv_sam;
			mi_fb_put(state, sp->reflect_buf, &sum_reflect);
		}
		if(sp->refract_pass)
		{
			sum_refract.r *= inv_sam;
			sum_refract.g *= inv_sam;
			sum_refract.b *= inv_sam;
			mi_fb_put(state, sp->refract_buf, &sum_refract);
		}
		if(sp->incandescence_pass)
		{
			sum_incandescence.r *= inv_sam;
			sum_incandescence.g *= inv_sam;
			sum_incandescence.b *= inv_sam;
			mi_fb_put(state, sp->incandescence_buf, &sum_incandescence);
		}
		if(sp->z_pass)
		{
			sum_z.r *= inv_sam;
			sum_z.g *= inv_sam;
			sum_z.b *= inv_sam;
			mi_fb_put(state, sp->z_buf, &sum_z);
		}
		if(sp->bent_normal_pass)
		{
			sum_bent_normal.r *= inv_sam;
			sum_bent_normal.g *= inv_sam;
			sum_bent_normal.b *= inv_sam;
			mi_fb_put(state, sp->bent_normal_buf, &sum_bent_normal);
		}
		if(sp->specialC1_pass)
		{
			sum_specialC1.r *= inv_sam;
			sum_specialC1.g *= inv_sam;
			sum_specialC1.b *= inv_sam;
			mi_fb_put(state, sp->specialC1_buf, &sum_specialC1);
		}
		if(sp->specialC2_pass)
		{
			sum_specialC2.r *= inv_sam;
			sum_specialC2.g *= inv_sam;
			sum_specialC2.b *= inv_sam;
			mi_fb_put(state, sp->specialC2_buf, &sum_specialC2);
		}
		if(sp->specialC3_pass)
		{
			sum_specialC3.r *= inv_sam;
			sum_specialC3.g *= inv_sam;
			sum_specialC3.b *= inv_sam;
			mi_fb_put(state, sp->specialC3_buf, &sum_specialC3);
		}
		if(sp->specialC4_pass)
		{
			sum_specialC4.r *= inv_sam;
			sum_specialC4.g *= inv_sam;
			sum_specialC4.b *= inv_sam;
			mi_fb_put(state, sp->specialC4_buf, &sum_specialC4);
		}
		if(sp->specialC5_pass)
		{
			sum_specialC5.r *= inv_sam;
			sum_specialC5.g *= inv_sam;
			sum_specialC5.b *= inv_sam;
			mi_fb_put(state, sp->specialC5_buf, &sum_specialC5);
		}
		if(sp->specialC6_pass)
		{
			sum_specialC6.r *= inv_sam;
			sum_specialC6.g *= inv_sam;
			sum_specialC6.b *= inv_sam;
			mi_fb_put(state, sp->specialC6_buf, &sum_specialC6);
		}
		if(sp->specialC7_pass)
		{
			sum_specialC7.r *= inv_sam;
			sum_specialC7.g *= inv_sam;
			sum_specialC7.b *= inv_sam;
			mi_fb_put(state, sp->specialC7_buf, &sum_specialC7);
		}
		if(sp->specialC8_pass)
		{
			sum_specialC8.r *= inv_sam;
			sum_specialC8.g *= inv_sam;
			sum_specialC8.b *= inv_sam;
			mi_fb_put(state, sp->specialC8_buf, &sum_specialC8);
		}
		if(sp->specialC9_pass)
		{
			sum_specialC9.r *= inv_sam;
			sum_specialC9.g *= inv_sam;
			sum_specialC9.b *= inv_sam;
			mi_fb_put(state, sp->specialC9_buf, &sum_specialC9);
		}
		if(sp->specialC10_pass)
		{
			sum_specialC10.r *= inv_sam;
			sum_specialC10.g *= inv_sam;
			sum_specialC10.b *= inv_sam;
			mi_fb_put(state, sp->specialC10_buf, &sum_specialC10);
		}
	}
	return(miTRUE);
}
