//  Copyright (c) 2007, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"
#include "geoshader.h"

typedef struct
{
	miBoolean	enable;
	miBoolean	verbose;
	miTag		material;
	int			i_object;
	int			n_object;
	miTag		object[1];

}p_shader_replacer_t;

extern "C"
{
DLLEXPORT int p_shader_replacer_version(void) {return(1);}

DLLEXPORT miBoolean p_shader_replacer(miTag *result, miState *state, p_shader_replacer_t *param)
{
	miBoolean enable = *mi_eval_boolean(&param->enable);
	miBoolean verbose = *mi_eval_boolean(&param->verbose);
	miTag material = *mi_eval_tag(&param->material);

//	miTag object = *mi_eval_tag(&param->object);
	miTag *object_pt = mi_eval_tag(param->object);
	int i_object = *mi_eval_integer(&param->i_object);
	int n_object = *mi_eval_integer(&param->n_object);

	//if shader enabled go on
	if (enable)
	{	
		if (verbose)
		{
			mi_info("*************************************\n");
			mi_info("    p_shader_replacer\n");
			mi_info("*************************************\n");
		}

		char *c_object;
		const char *obj;
		char *c_inst;
		miTag t_obj;

		int i;
		for (i=0; i < n_object; i++)
		{
			miTag object = miNULLTAG;
			object = object_pt[i_object + i];
//			mi_warning("eee %i", object);
			if (object) //check if named object exists in scene db, return object as a string
			{
				c_object = (char*)mi_db_access(object);
				int tagtype =  mi_db_type(object);
				if (tagtype == miSCENE_INSTANCE)	//check if it is an instance
				{
					c_inst = (char*)mi_api_tag_lookup(object);
					if (verbose)
						mi_info("p_shader_replacer->instance:: %s\n", c_inst);
					
					// get object that is being instanced (if any)	
					if(mi_query(miQ_INST_ITEM, NULL, object, &t_obj))
					{				
						obj = mi_api_tag_lookup(t_obj);				
						if (verbose)
							mi_info("p_shader_replacer->instance->object:: %s\n", obj);	
					}
					else
					{
						mi_error("p_shader_replacer: no INSTANCE_ITEM (shape) exist on instance \"%s\"", c_object);
						mi_db_unpin(object);
						return(miFALSE);
					}
				} 
				else
				{
					mi_error("p_shader_replacer: \"%s\" is not miSCENE_INSTANCE\n", c_object);
					mi_db_unpin(object);
					return(miFALSE);
				}
				mi_db_unpin(object);
			}
			else
			{
				mi_error("p_shader_replacer: no object with given TAG \"%i\" was found in scene DB. Or no object specified\n", object);
				return(miFALSE);
			}
			//get our materialSG

			char *c_material;	
			if (material) //check if named material exists in scene db, return it as a string
			{
				c_material = (char*)mi_db_access(material);
				int tagtype =  mi_db_type(material);
				if (tagtype == miSCENE_MATERIAL)	//check it is a material
				{
					const char *c_matSG = mi_api_tag_lookup(material);
					if (verbose)
						mi_info("p_shader_replacer->material:: %s\n", c_matSG);
				
					//get the shader of the SG
					miTag t_mat;
					mi_query(miQ_MTL_SHADER, NULL, material, &t_mat );
			
					const char *mat = mi_api_tag_lookup(t_mat);
				
					if (verbose)
						mi_info("p_shader_replacer->material->shader:: %s\n", mat);
				
					//get shadow of SG		
					miTag t_shadow;
					mi_query(miQ_MTL_SHADOW, NULL, material, &t_shadow );
			
					const char *shadow = mi_api_tag_lookup(t_shadow);
				
					if (verbose)
						mi_info("p_shader_replacer->material->shadow:: %s\n", shadow);				
				}
				else
				{
					mi_error("p_shader_replacer: \"%s\" is not miSCENE_MATERIAL\n", c_material);
					mi_db_unpin(material);
					return(miFALSE);
				}
				mi_db_unpin(material);		
			}
			else
			{
				mi_error("p_shader_replacer: no material with given TAG \"%i\" was found in scene DB. Or no material specified\n", material);
				return(miFALSE);
			}

			miTag t_omat;
			mi_query(miQ_INST_MATERIAL, NULL, object, &t_omat);

			const char *omat = mi_api_tag_lookup(t_omat);
			
			if (verbose)
				mi_info("p_shader_replacer->oldinstance->material:: %s\n", omat);
			
	//		miBoolean ooff;
	//		mi_query(miQ_INST_HIDE, NULL, object, &ooff );

	//		if (verbose == 1)
	//		{
	//			if (ooff == 0)
	//				mi_info("p_shader_replacer->oldinstance->off::miFALSE \n");
	//			else 
	//				mi_info("p_shader_replacer->oldinstance->off::miTRUE \n");			
	//		}

			//go incermaental -- currently not always working
			miTag instEnd;						
			if (verbose)
				mi_info("p_shader_replacer: should replace the old material on %s\n",c_inst );


			mi_api_incremental(1);
			miInstance *instance = mi_api_instance_begin(mi_mem_strdup(c_inst));
			
			instance->mtl_override = miTRUE;
			instance->mtl_array_size = 0; 
	//		instance->material = mi_phen_clone(state, material);
			instance->material = material;
	//		instance->off = 0;

			instEnd = mi_api_instance_end(mi_mem_strdup(obj), 0, -1);
			if (verbose)
				mi_info("*************************************\n");
		}  //for	
		if (verbose)
		{
			mi_info("p_shader_replacer: Finished\n");
			mi_info("*************************************\n");
		}
		return(miTRUE);		

	} //if enable
	else
		return(miTRUE);
	return(miTRUE);
}
}
