//  Copyright (c) 2005, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifdef HPUX
#pragma OPT_LEVEL 1	/* workaround for HP/UX optimizer bug, +O2 and +O3 */
#endif

//#include <stdio.h>
//#include <stdlib.h>		/* for abs */
//#include <float.h>		/* for FLT_MAX */
#include <math.h>
#include <string.h>
#include <assert.h>
#include "shader.h"

typedef struct
{
//	miBoolean	normalize;
	miBoolean self_shadow;
	int		mode;
	int		i_light;
	int		n_light;
	miTag	light[1];
}p_shadow_t;

extern "C"
{
DLLEXPORT int p_shadow_version(void) {return(2);}

DLLEXPORT miBoolean p_shadow(miColor *result, miState *state, p_shadow_t *paras)
{
	if(state->type == miRAY_SHADOW || state->type == miRAY_DISPLACE)
		return miFALSE;

	miTag		*light;
	int			n_l;
	int			i_l;
	int			m;
	int			n;
	int			samples;
	miColor		color;
	miColor		sum;
	miScalar	dot_nl;
	miVector	dir;

	m    = *mi_eval_integer(&paras->mode);
	n_l   = *mi_eval_integer(&paras->n_light);
	i_l   = *mi_eval_integer(&paras->i_light);
	light =  mi_eval_tag(paras->light) + i_l;
	miBoolean self_shd = *mi_eval_boolean(&paras->self_shadow);

	miVector orig_normal = state->normal;		//copy original normal
	miOptions *orig_options = (miOptions *)(state->options);	//save pointer to original options
	miOptions m_options = *state->options;		//copy original options
	state->options = &m_options;				//link new options to state

	if (m == 1)
		mi_inclusive_lightlist(&n_l, &light, state);
	else if (m == 2)
		mi_exclusive_lightlist(&n_l, &light, state);

	result->r = result->g = result->b = result->a = 0;

	for (n=0; n < n_l; n++, light++)
	{
		sum.r = sum.g = sum.b = 0;
		samples = 0;
		while (mi_sample_light(&color, &dir, &dot_nl, state, *light, &samples))
		{
			sum.r += color.r;
			sum.g += color.g;
			sum.b += color.b;
		}
		if (samples)
		{
			result->r += sum.r / samples;
			result->g += sum.g / samples;
			result->b += sum.b / samples;
		}
		
		if(self_shd != miTRUE)							//calculate lighting one more time but without shadow
		{
			state->normal.x = state->normal.y = state->normal.z = 0;
			m_options.shadow = 0;					//edit new options, disable shadow
			sum.r = sum.g = sum.b = 0;
			samples = 0;
			while (mi_sample_light(&color, &dir, &dot_nl, state, *light, &samples))
			{
				dot_nl = mi_vector_dot(&orig_normal, &dir);
				if(dot_nl <= 0)							//add reslt only for backside
				{
					sum.r += color.r;
					sum.g += color.g;
					sum.b += color.b;
				}
			}
			if(samples)
			{
				result->r += sum.r / samples;
				result->g += sum.g / samples;
				result->b += sum.b / samples;
			}
			m_options.shadow = orig_options->shadow;		//restore original shadow
			state->normal = orig_normal;						//restore original normal
		}
	}
	state->options = orig_options;				//restore pointer to original options
	return(miTRUE);
}
}
